# -*- coding: utf-8 -*-
import pytest
from mock import patch, MagicMock

from awsscripter.common.connection_manager import ConnectionManager
from awsscripter.common.exceptions import InvalidHookArgumentSyntaxError
from awsscripter.common.exceptions import InvalidHookArgumentTypeError
from awsscripter.common.exceptions import InvalidHookArgumentValueError
from awsscripter.hooks.asg_scaling_processes import ASGScalingProcesses
from awsscripter.stack import Stack


class TestASGScalingProcesses(object):
    def setup_method(self, test_method):
        self.stack = MagicMock(spec=Stack)
        self.stack.connection_manager = MagicMock(spec=ConnectionManager)
        self.stack.external_name = "external_name"
        self.asg_scaling_processes = ASGScalingProcesses(None, self.stack)

    def test_get_stack_resources_sends_correct_request(self):
        self.stack.connection_manager.call.return_value = {
            "StackResources": [
                {
                    "ResourceType": "AWS::AutoScaling::AutoScalingGroup",
                    'PhysicalResourceId': 'cloudreach-examples-asg'
                }
            ]
        }
        self.asg_scaling_processes._get_stack_resources()
        self.stack.connection_manager.call.assert_called_with(
                service="cloudformation",
                command="describe_stack_resources",
                kwargs={
                    "StackName": "external_name",
                }
            )

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._get_stack_resources"
    )
    def test_find_autoscaling_groups_with_stack_with_asgs(
        self, mock_get_stack_resources
    ):
        mock_get_stack_resources.return_value = [{
            'LogicalResourceId': 'AutoScalingGroup',
            'PhysicalResourceId': 'cloudreach-examples-asg',
            'ResourceStatus': 'CREATE_COMPLETE',
            'ResourceType': 'AWS::AutoScaling::AutoScalingGroup',
            'StackId': 'arn:aws:...',
            'StackName': 'cloudreach-examples-dev-vpc'
        }]
        response = self.asg_scaling_processes._find_autoscaling_groups()

        assert response == ["cloudreach-examples-asg"]

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._get_stack_resources"
    )
    def test_find_autoscaling_groups_with_stack_without_asgs(
        self, mock_get_stack_resources
    ):
        mock_get_stack_resources.return_value = []
        response = self.asg_scaling_processes._find_autoscaling_groups()

        assert response == []

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._find_autoscaling_groups"
    )
    def test_run_with_resume_argument(self, mock_find_autoscaling_groups):
        self.asg_scaling_processes.argument = u"resume::ScheduledActions"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        self.asg_scaling_processes.run()
        self.stack.connection_manager.call.assert_called_once_with(
                service="autoscaling",
                command="resume_processes",
                kwargs={
                    "AutoScalingGroupName": "autoscaling_group_1",
                    "ScalingProcesses": [
                        "ScheduledActions"
                    ]
                }
            )

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._find_autoscaling_groups"
    )
    def test_run_with_suspend_argument(self, mock_find_autoscaling_groups):
        self.asg_scaling_processes.argument = "suspend::ScheduledActions"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        self.asg_scaling_processes.run()
        self.stack.connection_manager.call.assert_called_once_with(
                service="autoscaling",
                command="suspend_processes",
                kwargs={
                    "AutoScalingGroupName": "autoscaling_group_1",
                    "ScalingProcesses": [
                        "ScheduledActions"
                    ]
                }
            )

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._find_autoscaling_groups"
    )
    def test_run_with_invalid_string_argument(
        self, mock_find_autoscaling_groups
    ):
        self.asg_scaling_processes.argument = u"invalid_string"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        with pytest.raises(InvalidHookArgumentSyntaxError):
            self.asg_scaling_processes.run()

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._find_autoscaling_groups"
    )
    def test_run_with_unsupported_argument(self, mock_find_autoscaling_groups):
        self.asg_scaling_processes.argument = "start::Healthcheck"
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        with pytest.raises(InvalidHookArgumentValueError):
            self.asg_scaling_processes.run()

    @patch(
        "awsscripter.hooks.asg_scaling_processes"
        ".ASGScalingProcesses._find_autoscaling_groups"
    )
    def test_run_with_non_string_argument(self, mock_find_autoscaling_groups):
        self.asg_scaling_processes.argument = 10
        mock_find_autoscaling_groups.return_value = ["autoscaling_group_1"]
        with pytest.raises(InvalidHookArgumentTypeError):
            self.asg_scaling_processes.run()



import click

from awsscripter.stack.helpers import catch_exceptions, get_stack_or_env, write
from awsscripter.stack.helpers import simplify_change_set_description


@click.group(name="describe")
@click.pass_context
def describe_group(ctx):
    """
    Commands for describing attributes of stacks.
    """
    pass


@describe_group.command(name="change-set")
@click.argument("path")
@click.argument("change-set-name")
@click.option(
    "-v", "--verbose", is_flag=True, help="Display verbose output."
)
@click.pass_context
@catch_exceptions
def describe_change_set(ctx, path, change_set_name, verbose):
    """
    Describes the change set.

    """
    stack, _ = get_stack_or_env(ctx, path)
    description = stack.describe_change_set(change_set_name)
    if not verbose:
        description = simplify_change_set_description(description)
    write(description, ctx.obj["output_format"])


@describe_group.command(name="policy")
@click.argument("path")
@click.pass_context
@catch_exceptions
def describe_policy(ctx, path):
    """
    Displays the stack policy used.

    """
    stack, _ = get_stack_or_env(ctx, path)
    response = stack.get_policy()
    write(response.get('StackPolicyBody', {}))



import logging
import os
import sys
from functools import wraps
from json import JSONEncoder

import click
import yaml
from boto3.exceptions import Boto3Error
from botocore.exceptions import BotoCoreError, ClientError
from jinja2.exceptions import TemplateError

from awsscripter.common.config_reader import ConfigReader
from awsscripter.common.exceptions import awsscripterException
from awsscripter.stack.stack_status_colourer import StackStatusColourer


def catch_exceptions(func):
    """
    Catches and simplifies expected errors thrown by awsscripter.

    catch_exceptions should be used as a decorator.

    :param func: The function which may throw exceptions which should be
        simplified.
    :type func: func
    :returns: The decorated function.
    :rtype: func
    """
    @wraps(func)
    def decorated(*args, **kwargs):
        """
        Invokes ``func``, catches expected errors, prints the error message and
        exits awsscripter with a non-zero exit code.
        """
        try:
            return func(*args, **kwargs)
        except (awsscripterException, BotoCoreError, ClientError, Boto3Error,
                TemplateError) as error:
            write(error)
            sys.exit(1)

    return decorated


def confirmation(
    command, ignore, environment=None, stack=None, change_set=None
):
    if not ignore:
        msg = "Do you want to {} ".format(command)
        if environment:
            msg = msg + "environment '{0}'?".format(environment)
        elif change_set and stack:
            msg = msg + "change set '{0}' for stack '{1}'".format(
                change_set, stack
            )
        elif stack:
            msg = msg + "stack '{0}'".format(stack)
        click.confirm(msg, abort=True)


def write(var, output_format="str", no_colour=True):
    """
    Writes ``var`` to stdout. If output_format is set to "json" or "yaml",
    write ``var`` as a JSON or YAML string.

    :param var: The object to print
    :type var: obj
    :param output_format: The format to print the output as. Allowed values: \
    "str", "json", "yaml"
    :type output_format: str
    :param no_colour: Whether to colour stack statuses
    :type no_colour: bool
    """
    if output_format == "json":
        encoder = CustomJsonEncoder()
        stream = encoder.encode(var)
    if output_format == "yaml":
        stream = yaml.safe_dump(var, default_flow_style=False)
    if output_format == "str":
        stream = var

    if not no_colour:
        stack_status_colourer = StackStatusColourer()
        stream = stack_status_colourer.colour(stream)

    click.echo(stream)


def get_stack_or_env(ctx, path):
    """
    Parses the path to generate relevant Environment and Stack object.

    :param ctx: Cli context.
    :type ctx: click.Context
    :param path: Path to either stack config or environment folder.
    :type path: str
    """
    stack = None
    env = None

    config_reader = ConfigReader(
        ctx.obj["awsscripter_dir"], ctx.obj["user_variables"]
    )

    if os.path.splitext(path)[1]:
        stack = config_reader.construct_stack(path)
    else:
        env = config_reader.construct_environment(path)

    return (stack, env)


def setup_logging(debug, no_colour):
    """
    Sets up logging.

    By default, the python logging module is configured to push logs to stdout
    as long as their level is at least INFO. The log format is set to
    "[%(asctime)s] - %(name)s - %(message)s" and the date format is set to
    "%Y-%m-%d %H:%M:%S".

    After this function has run, modules should:

    .. code:: python

        import logging

        logging.getLogger(__name__).info("my log message")

    :param debug: A flag indication whether to turn on debug logging.
    :type debug: bool
    :no_colour: A flag to indicating whether to turn off coloured output.
    :type no_colour: bool
    :returns: A logger.
    :rtype: logging.Logger
    """
    if debug:
        awsscripter_logging_level = logging.DEBUG
        logging.getLogger("botocore").setLevel(logging.INFO)
    else:
        awsscripter_logging_level = logging.INFO
        # Silence botocore logs
        logging.getLogger("botocore").setLevel(logging.CRITICAL)

    formatter_class = logging.Formatter if no_colour else ColouredFormatter

    formatter = formatter_class(
        fmt="[%(asctime)s] - %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S"
    )

    log_handler = logging.StreamHandler()
    log_handler.setFormatter(formatter)
    logger = logging.getLogger("awsscripter")
    logger.addHandler(log_handler)
    logger.setLevel(awsscripter_logging_level)
    return logger


def simplify_change_set_description(response):
    """
    Simplies the response from the AWS describe change set API.

    :param response: The original api response.
    :type response: dict
    :returns: A more concise description of the change set.
    :rtype: dict
    """
    desired_response_items = [
        "ChangeSetName",
        "CreationTime",
        "ExecutionStatus",
        "StackName",
        "Status",
        "StatusReason"
    ]
    desired_resource_changes = [
        "Action",
        "LogicalResourceId",
        "PhysicalResourceId",
        "Replacement",
        "ResourceType",
        "Scope"
    ]
    formatted_response = {
        k: v
        for k, v in response.items()
        if k in desired_response_items
    }
    formatted_response["Changes"] = [
        {
            "ResourceChange": {
                k: v
                for k, v in change["ResourceChange"].items()
                if k in desired_resource_changes
            }
        }
        for change in response["Changes"]
    ]
    return formatted_response


class ColouredFormatter(logging.Formatter):
    """
    ColouredFormatter add colours to all stack statuses that appear in log
    messages.
    """

    stack_status_colourer = StackStatusColourer()

    def format(self, record):
        """
        Colours and returns all stack statuses in ``record``.

        :param record: The log item to format.
        :type record: str
        :returns: str
        """
        response = super(ColouredFormatter, self).format(record)
        coloured_response = self.stack_status_colourer.colour(response)
        return coloured_response


class CustomJsonEncoder(JSONEncoder):
    """
    CustomJsonEncoder is a JSONEncoder which encodes all items as JSON by
    calling their __str__() method.
    """

    def default(self, item):
        """
        Returns stringified version of item.

        :param item: An arbitrary object to stringify.
        :type item: obj
        :returns: The stringified object.
        :rtype: str
        """
        return str(item)



# -*- coding: utf-8 -*-

from troposphere import Template, Parameter, Ref, Output

from troposphere.ec2 import VPC, InternetGateway, VPCGatewayAttachment


class VpcTemplate(object):

    def __init__(self):
        self.template = Template()

        self.add_parameters()

        self.add_vpc()
        self.add_igw()

        self.add_outputs()

    def add_parameters(self):
        t = self.template

        self.cidr_block_param = t.add_parameter(Parameter(
            "CidrBlock",
            Type="String",
            Default="10.0.0.0/16",
        ))

    def add_vpc(self):
        t = self.template

        self.vpc = t.add_resource(VPC(
            "VirtualPrivateCloud",
            CidrBlock=Ref(self.cidr_block_param),
            InstanceTenancy="default",
            EnableDnsSupport=True,
            EnableDnsHostnames=True,
        ))

    def add_igw(self):
        t = self.template

        self.igw = t.add_resource(InternetGateway(
            "InternetGateway",
        ))

        t.add_resource(VPCGatewayAttachment(
            "IGWAttachment",
            VpcId=Ref(self.vpc),
            InternetGatewayId=Ref(self.igw),
        ))

    def add_outputs(self):
        t = self.template

        t.add_output(Output(
            "VpcId",
            Description="New VPC ID",
            Value=Ref(self.vpc)
        ))


def sceptre_handler(sceptre_user_data):
    vpc = VpcTemplate()
    return vpc.template.to_json()



# -*- coding: utf-8 -*-

import datetime

import pytest
from botocore.exceptions import ClientError
from dateutil.tz import tzutc
from mock import patch, sentinel, MagicMock, Mock

from awsscripter.common.exceptions import CannotUpdateFailedStackError
from awsscripter.common.exceptions import ProtectedStackError
from awsscripter.common.exceptions import StackDoesNotExistError
from awsscripter.common.exceptions import UnknownStackChangeSetStatusError
from awsscripter.common.exceptions import UnknownStackStatusError
from awsscripter.common.template import Template
from awsscripter.stack import Stack
from awsscripter.stack.stack_status import StackChangeSetStatus
from awsscripter.stack.stack_status import StackStatus


class TestStack(object):

    def setup_method(self, test_method):
        self.patcher_connection_manager = patch(
            "awsscripter.stack.ConnectionManager"
        )
        self.mock_ConnectionManager = self.patcher_connection_manager.start()
        self.stack = Stack(
            name=sentinel.stack_name, project_code=sentinel.project_code,
            template_path=sentinel.template_path, region=sentinel.region,
            iam_role=sentinel.iam_role, parameters={"key1": "val1"},
            sceptre_user_data=sentinel.sceptre_user_data, hooks={},
            s3_details=None, dependencies=sentinel.dependencies,
            role_arn=sentinel.role_arn, protected=False,
            tags={"tag1": "val1"}, external_name=sentinel.external_name,
            notifications=[sentinel.notification],
            on_failure=sentinel.on_failure
        )
        self.stack._template = MagicMock(spec=Template)

    def teardown_method(self, test_method):
        self.patcher_connection_manager.stop()

    def test_initiate_stack(self):
        stack = Stack(
            name=sentinel.stack_name, project_code=sentinel.project_code,
            template_path=sentinel.template_path, region=sentinel.region,
            external_name=sentinel.external_name
        )
        self.mock_ConnectionManager.assert_called_with(
            sentinel.region, None
        )
        assert stack.name == sentinel.stack_name
        assert stack.project_code == sentinel.project_code
        assert stack.external_name == sentinel.external_name
        assert stack.hooks == {}
        assert stack.parameters == {}
        assert stack.sceptre_user_data == {}
        assert stack.template_path == sentinel.template_path
        assert stack.s3_details is None
        assert stack._template is None
        assert stack.protected is False
        assert stack.role_arn is None
        assert stack.dependencies == []
        assert stack.tags == {}
        assert stack.notifications == []
        assert stack.on_failure is None

    def test_repr(self):
        self.stack.connection_manager.region = sentinel.region
        self.stack.connection_manager.iam_role = None

        assert self.stack.__repr__() == \
            "awsscripter.stack.Stack(" \
            "name='sentinel.stack_name', " \
            "project_code='sentinel.project_code', " \
            "template_path='sentinel.template_path', " \
            "region='sentinel.region', " \
            "iam_role='None', parameters='{'key1': 'val1'}', " \
            "sceptre_user_data='sentinel.sceptre_user_data', " \
            "hooks='{}', s3_details='None', " \
            "dependencies='sentinel.dependencies', "\
            "role_arn='sentinel.role_arn', " \
            "protected='False', tags='{'tag1': 'val1'}', " \
            "external_name='sentinel.external_name', " \
            "notifications='[sentinel.notification]', " \
            "on_failure='sentinel.on_failure'" \
            ")"

    @patch("awsscripter.stack.Template")
    def test_template_loads_template(self, mock_Template):
        self.stack._template = None
        mock_Template.return_value = sentinel.template
        response = self.stack.template

        mock_Template.assert_called_once_with(
            path=sentinel.template_path,
            sceptre_user_data=sentinel.sceptre_user_data,
            connection_manager=self.stack.connection_manager,
            s3_details=None
        )
        assert response == sentinel.template

    def test_template_returns_template_if_it_exists(self):
        self.stack._template = sentinel.template
        response = self.stack.template
        assert response == sentinel.template

    def test_external_name_with_custom_stack_name(self):
        stack = Stack(
            name="stack_name", project_code="project_code",
            template_path="template_path", region="region",
            external_name="external_name"
        )

        assert stack.external_name == "external_name"

    @patch("awsscripter.stack.Stack._wait_for_completion")
    def test_create_sends_correct_request(self, mock_wait_for_completion):
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }

        self.stack.create()

        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ],
                "OnFailure": sentinel.on_failure
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("awsscripter.stack.Stack._wait_for_completion")
    def test_create_sends_correct_request_no_notifications(
        self, mock_wait_for_completion
    ):
        self.stack._template = Mock(spec=Template)
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        self.stack.notifications = []

        self.stack.create()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ],
                "OnFailure": sentinel.on_failure
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("awsscripter.stack.Stack._wait_for_completion")
    def test_create_sends_correct_request_with_no_failure(
        self, mock_wait_for_completion
    ):
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        self.stack.on_failure = None

        self.stack.create()

        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("awsscripter.stack.Stack._wait_for_completion")
    def test_update_sends_correct_request(self, mock_wait_for_completion):
        self.stack._template = Mock(spec=Template)
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }

        self.stack.update()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="update_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("awsscripter.stack.Stack._wait_for_completion")
    def test_update_sends_correct_request_no_notification(
            self, mock_wait_for_completion
    ):
        self.stack._template = Mock(spec=Template)
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }

        self.stack.notifications = []
        self.stack.update()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="update_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    @patch("awsscripter.stack.Stack.create")
    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_stack_that_does_not_exist(
            self, mock_get_status, mock_create
    ):
        mock_get_status.side_effect = StackDoesNotExistError()
        mock_create.return_value = sentinel.launch_response
        response = self.stack.launch()
        mock_create.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("awsscripter.stack.Stack.create")
    @patch("awsscripter.stack.Stack.delete")
    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_stack_that_failed_to_create(
            self, mock_get_status, mock_delete, mock_create
    ):
        mock_get_status.return_value = "CREATE_FAILED"
        mock_create.return_value = sentinel.launch_response
        response = self.stack.launch()
        mock_delete.assert_called_once_with()
        mock_create.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("awsscripter.stack.Stack.update")
    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_complete_stack_with_updates_to_perform(
            self, mock_get_status, mock_update
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.return_value = sentinel.launch_response
        response = self.stack.launch()
        mock_update.assert_called_once_with()
        assert response == sentinel.launch_response

    @patch("awsscripter.stack.Stack.update")
    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_complete_stack_with_no_updates_to_perform(
            self, mock_get_status, mock_update
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.side_effect = ClientError(
            {
                "Error": {
                    "Code": "NoUpdateToPerformError",
                    "Message": "No updates are to be performed."
                }
            },
            sentinel.operation
        )
        response = self.stack.launch()
        mock_update.assert_called_once_with()
        assert response == StackStatus.COMPLETE

    @patch("awsscripter.stack.Stack.update")
    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_complete_stack_with_unknown_client_error(
            self, mock_get_status, mock_update
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_update.side_effect = ClientError(
            {
                "Error": {
                    "Code": "Boom!",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack.launch()

    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_in_progress_stack(self, mock_get_status):
        mock_get_status.return_value = "CREATE_IN_PROGRESS"
        response = self.stack.launch()
        assert response == StackStatus.IN_PROGRESS

    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_failed_stack(self, mock_get_status):
        mock_get_status.return_value = "UPDATE_FAILED"
        with pytest.raises(CannotUpdateFailedStackError):
            response = self.stack.launch()
            assert response == StackStatus.FAILED

    @patch("awsscripter.stack.Stack.get_status")
    def test_launch_with_unknown_stack_status(self, mock_get_status):
        mock_get_status.return_value = "UNKNOWN_STATUS"
        with pytest.raises(UnknownStackStatusError):
            self.stack.launch()

    @patch("awsscripter.stack.Stack._wait_for_completion")
    @patch("awsscripter.stack.Stack.get_status")
    def test_delete_with_created_stack(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"

        self.stack.delete()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="delete_stack",
            kwargs={
                "StackName": sentinel.external_name,
                "RoleARN": sentinel.role_arn
            }
        )

    @patch("awsscripter.stack.Stack._wait_for_completion")
    @patch("awsscripter.stack.Stack.get_status")
    def test_delete_when_wait_for_completion_raises_stack_does_not_exist_error(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_wait_for_completion.side_effect = StackDoesNotExistError()
        status = self.stack.delete()
        assert status == StackStatus.COMPLETE

    @patch("awsscripter.stack.Stack._wait_for_completion")
    @patch("awsscripter.stack.Stack.get_status")
    def test_delete_when_wait_for_completion_raises_non_existent_client_error(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Stack does not exist"
                }
            },
            sentinel.operation
        )
        status = self.stack.delete()
        assert status == StackStatus.COMPLETE

    @patch("awsscripter.stack.Stack._wait_for_completion")
    @patch("awsscripter.stack.Stack.get_status")
    def test_delete_when_wait_for_completion_raises_unexpected_client_error(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.return_value = "CREATE_COMPLETE"
        mock_wait_for_completion.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Boom"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack.delete()

    @patch("awsscripter.stack.Stack._wait_for_completion")
    @patch("awsscripter.stack.Stack.get_status")
    def test_delete_with_non_existent_stack(
            self, mock_get_status, mock_wait_for_completion
    ):
        mock_get_status.side_effect = StackDoesNotExistError()
        status = self.stack.delete()
        assert status == StackStatus.COMPLETE

    def test_describe_stack_sends_correct_request(self):
        self.stack.describe()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stacks",
            kwargs={"StackName": sentinel.external_name}
        )

    def test_describe_events_sends_correct_request(self):
        self.stack.describe_events()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stack_events",
            kwargs={"StackName": sentinel.external_name}
        )

    def test_describe_resources_sends_correct_request(self):
        self.stack.connection_manager.call.return_value = {
            "StackResources": [
                {
                    "LogicalResourceId": sentinel.logical_resource_id,
                    "PhysicalResourceId": sentinel.physical_resource_id,
                    "OtherParam": sentinel.other_param
                }
            ]
        }
        response = self.stack.describe_resources()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_stack_resources",
            kwargs={"StackName": sentinel.external_name}
        )
        assert response == [
            {
                "LogicalResourceId": sentinel.logical_resource_id,
                "PhysicalResourceId": sentinel.physical_resource_id
            }
        ]

    @patch("awsscripter.stack.Stack.describe")
    def test_describe_outputs_sends_correct_request(self, mock_describe):
        mock_describe.return_value = {
            "Stacks": [{
                "Outputs": sentinel.outputs
            }]
        }
        response = self.stack.describe_outputs()
        mock_describe.assert_called_once_with()
        assert response == sentinel.outputs

    @patch("awsscripter.stack.Stack.describe")
    def test_describe_outputs_handles_stack_with_no_outputs(
            self, mock_describe
    ):
        mock_describe.return_value = {
            "Stacks": [{}]
        }
        response = self.stack.describe_outputs()
        assert response == []

    def test_continue_update_rollback_sends_correct_request(self):
        self.stack.continue_update_rollback()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="continue_update_rollback",
            kwargs={
                "StackName": sentinel.external_name,
                "RoleARN": sentinel.role_arn
            }
        )

    def test_set_stack_policy_sends_correct_request(self):
        self.stack.set_policy("tests/fixtures/stack_policies/unlock.json")
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="set_stack_policy",
            kwargs={
                "StackName": sentinel.external_name,
                "StackPolicyBody": """{
  "Statement" : [
    {
      "Effect" : "Allow",
      "Action" : "Update:*",
      "Principal": "*",
      "Resource" : "*"
    }
  ]
}
"""
            }
        )

    def test_get_stack_policy_sends_correct_request(self):
        self.stack.get_policy()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="get_stack_policy",
            kwargs={
                "StackName": sentinel.external_name
            }
        )

    def test_create_change_set_sends_correct_request(self):
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }

        self.stack.create_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_change_set",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "ChangeSetName": sentinel.change_set_name,
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [sentinel.notification],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )

    def test_create_change_set_sends_correct_request_no_notifications(self):
        self.stack._template = Mock(spec=Template)
        self.stack._template.get_boto_call_parameter.return_value = {
            "Template": sentinel.template
        }
        self.stack.notifications = []

        self.stack.create_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="create_change_set",
            kwargs={
                "StackName": sentinel.external_name,
                "Template": sentinel.template,
                "Parameters": [{
                    "ParameterKey": "key1",
                    "ParameterValue": "val1"
                }],
                "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                "ChangeSetName": sentinel.change_set_name,
                "RoleARN": sentinel.role_arn,
                "NotificationARNs": [],
                "Tags": [
                    {"Key": "tag1", "Value": "val1"}
                ]
            }
        )

    def test_delete_change_set_sends_correct_request(self):
        self.stack.delete_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="delete_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )

    def test_describe_change_set_sends_correct_request(self):
        self.stack.describe_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="describe_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )

    @patch("awsscripter.stack.Stack._wait_for_completion")
    def test_execute_change_set_sends_correct_request(
        self, mock_wait_for_completion
    ):
        self.stack.execute_change_set(sentinel.change_set_name)
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="execute_change_set",
            kwargs={
                "ChangeSetName": sentinel.change_set_name,
                "StackName": sentinel.external_name
            }
        )
        mock_wait_for_completion.assert_called_once_with()

    def test_list_change_sets_sends_correct_request(self):
        self.stack.list_change_sets()
        self.stack.connection_manager.call.assert_called_with(
            service="cloudformation",
            command="list_change_sets",
            kwargs={"StackName": sentinel.external_name}
        )

    @patch("awsscripter.stack.Stack.set_policy")
    @patch("os.path.join")
    def test_lock_calls_set_stack_policy_with_policy(
            self, mock_join, mock_set_policy
    ):
        mock_join.return_value = "tests/fixtures/stack_policies/lock.json"
        self.stack.lock()
        mock_set_policy.assert_called_once_with(
            "tests/fixtures/stack_policies/lock.json"
        )

    @patch("awsscripter.stack.Stack.set_policy")
    @patch("os.path.join")
    def test_unlock_calls_set_stack_policy_with_policy(
            self, mock_join, mock_set_policy
    ):
        mock_join.return_value = "tests/fixtures/stack_policies/unlock.json"
        self.stack.unlock()
        mock_set_policy.assert_called_once_with(
            "tests/fixtures/stack_policies/unlock.json"
        )

    def test_format_parameters_with_sting_values(self):
        parameters = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3"
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1"},
            {"ParameterKey": "key2", "ParameterValue": "value2"},
            {"ParameterKey": "key3", "ParameterValue": "value3"}
        ]

    def test_format_parameters_with_none_values(self):
        parameters = {
            "key1": None,
            "key2": None,
            "key3": None
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == []

    def test_format_parameters_with_none_and_string_values(self):
        parameters = {
            "key1": "value1",
            "key2": None,
            "key3": "value3"
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1"},
            {"ParameterKey": "key3", "ParameterValue": "value3"}
        ]

    def test_format_parameters_with_list_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": ["value4", "value5", "value6"],
            "key3": ["value7", "value8", "value9"]
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4,value5,value6"},
            {"ParameterKey": "key3", "ParameterValue": "value7,value8,value9"}
        ]

    def test_format_parameters_with_none_and_list_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": None,
            "key3": ["value7", "value8", "value9"]
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key3", "ParameterValue": "value7,value8,value9"}
        ]

    def test_format_parameters_with_list_and_string_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": "value4",
            "key3": ["value5", "value6", "value7"]
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4"},
            {"ParameterKey": "key3", "ParameterValue": "value5,value6,value7"}
        ]

    def test_format_parameters_with_none_list_and_string_values(self):
        parameters = {
            "key1": ["value1", "value2", "value3"],
            "key2": "value4",
            "key3": None
        }
        formatted_parameters = self.stack._format_parameters(parameters)
        sorted_formatted_parameters = sorted(
            formatted_parameters,
            key=lambda x: x["ParameterKey"]
        )
        assert sorted_formatted_parameters == [
            {"ParameterKey": "key1", "ParameterValue": "value1,value2,value3"},
            {"ParameterKey": "key2", "ParameterValue": "value4"},
        ]

    @patch("awsscripter.stack.Stack.describe")
    def test_get_status_with_created_stack(self, mock_describe):
        mock_describe.return_value = {
            "Stacks": [{"StackStatus": "CREATE_COMPLETE"}]
        }
        status = self.stack.get_status()
        assert status == "CREATE_COMPLETE"

    @patch("awsscripter.stack.Stack.describe")
    def test_get_status_with_non_existent_stack(self, mock_describe):
        mock_describe.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Stack does not exist"
                }
            },
            sentinel.operation
        )
        with pytest.raises(StackDoesNotExistError):
            self.stack.get_status()

    @patch("awsscripter.stack.Stack.describe")
    def test_get_status_with_unknown_clinet_error(self, mock_describe):
        mock_describe.side_effect = ClientError(
            {
                "Error": {
                    "Code": "DoesNotExistException",
                    "Message": "Boom!"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack.get_status()

    def test_get_role_arn_without_role(self):
        self.stack.role_arn = None
        assert self.stack._get_role_arn() == {}

    def test_get_role_arn_with_role(self):
        assert self.stack._get_role_arn() == {"RoleARN": sentinel.role_arn}

    def test_protect_execution_without_protection(self):
        # Function should do nothing if protect == False
        self.stack._protect_execution()

    def test_protect_execution_without_explicit_protection(self):
        self.stack._protect_execution()

    def test_protect_execution_with_protection(self):
        self.stack.protected = True
        with pytest.raises(ProtectedStackError):
            self.stack._protect_execution()

    @patch("awsscripter.stack.time")
    @patch("awsscripter.stack.Stack._log_new_events")
    @patch("awsscripter.stack.Stack.get_status")
    @patch("awsscripter.stack.Stack._get_simplified_status")
    def test_wait_for_completion_calls_log_new_events(
            self, mock_get_simplified_status, mock_get_status,
            mock_log_new_events, mock_time
    ):
        mock_get_simplified_status.return_value = StackStatus.COMPLETE

        self.stack._wait_for_completion()
        mock_log_new_events.assert_called_once_with()

    @pytest.mark.parametrize("test_input,expected", [
        ("ROLLBACK_COMPLETE", StackStatus.FAILED),
        ("STACK_COMPLETE", StackStatus.COMPLETE),
        ("STACK_IN_PROGRESS", StackStatus.IN_PROGRESS),
        ("STACK_FAILED", StackStatus.FAILED)
    ])
    def test_get_simplified_status_with_known_stack_statuses(
            self, test_input, expected
    ):
        response = self.stack._get_simplified_status(test_input)
        assert response == expected

    def test_get_simplified_status_with_stack_in_unknown_state(self):
        with pytest.raises(UnknownStackStatusError):
            self.stack._get_simplified_status("UNKOWN_STATUS")

    @patch("awsscripter.stack.Stack.describe_events")
    def test_log_new_events_calls_describe_events(self, mock_describe_events):
        mock_describe_events.return_value = {
            "StackEvents": []
        }
        self.stack._log_new_events()
        self.stack.describe_events.assert_called_once_with()

    @patch("awsscripter.stack.Stack.describe_events")
    def test_log_new_events_prints_correct_event(self, mock_describe_events):
        self.stack.name = "stack-name"
        mock_describe_events.return_value = {
            "StackEvents": [
                {
                    "Timestamp": datetime.datetime(
                        2016, 3, 15, 14, 2, 0, 0, tzinfo=tzutc()
                    ),
                    "LogicalResourceId": "id-2",
                    "ResourceType": "type-2",
                    "ResourceStatus": "resource-status"
                },
                {
                    "Timestamp": datetime.datetime(
                        2016, 3, 15, 14, 1, 0, 0, tzinfo=tzutc()
                    ),
                    "LogicalResourceId": "id-1",
                    "ResourceType": "type-1",
                    "ResourceStatus": "resource",
                    "ResourceStatusReason": "User Initiated"
                }
            ]
        }
        self.stack.most_recent_event_datetime = (
            datetime.datetime(2016, 3, 15, 14, 0, 0, 0, tzinfo=tzutc())
        )
        self.stack._log_new_events()

    @patch("awsscripter.stack.time")
    @patch("awsscripter.stack.Stack._get_cs_status")
    def test_wait_for_cs_completion_calls_get_cs_status(
        self, mock_get_cs_status, mock_time
    ):
        mock_get_cs_status.side_effect = [
            StackChangeSetStatus.PENDING, StackChangeSetStatus.READY
        ]

        self.stack.wait_for_cs_completion(sentinel.change_set_name)
        mock_get_cs_status.assert_called_with(sentinel.change_set_name)

    @patch("awsscripter.stack.Stack.describe_change_set")
    def test_get_cs_status_handles_all_statuses(
        self, mock_describe_change_set
    ):
        scss = StackChangeSetStatus
        return_values = {                                                                                                     # NOQA
                 "Status":    ('CREATE_PENDING', 'CREATE_IN_PROGRESS', 'CREATE_COMPLETE', 'DELETE_COMPLETE', 'FAILED'),       # NOQA
        "ExecutionStatus": {                                                                                                  # NOQA
        'UNAVAILABLE':         (scss.PENDING,     scss.PENDING,         scss.PENDING,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'AVAILABLE':           (scss.PENDING,     scss.PENDING,         scss.READY,        scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_IN_PROGRESS': (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_COMPLETE':    (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'EXECUTE_FAILED':      (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        'OBSOLETE':            (scss.DEFUNCT,     scss.DEFUNCT,         scss.DEFUNCT,      scss.DEFUNCT,      scss.DEFUNCT),  # NOQA
        }                                                                                                                     # NOQA
        }                                                                                                                     # NOQA

        for i, status in enumerate(return_values['Status']):
            for exec_status, returns in \
                    return_values['ExecutionStatus'].items():
                mock_describe_change_set.return_value = {
                    "Status": status,
                    "ExecutionStatus": exec_status
                }
                response = self.stack._get_cs_status(sentinel.change_set_name)
                assert response == returns[i]

        for status in return_values['Status']:
            mock_describe_change_set.return_value = {
                "Status": status,
                "ExecutionStatus": 'UNKOWN_STATUS'
            }
            with pytest.raises(UnknownStackChangeSetStatusError):
                self.stack._get_cs_status(sentinel.change_set_name)

        for exec_status in return_values['ExecutionStatus'].keys():
            mock_describe_change_set.return_value = {
                "Status": 'UNKOWN_STATUS',
                "ExecutionStatus": exec_status
            }
            with pytest.raises(UnknownStackChangeSetStatusError):
                self.stack._get_cs_status(sentinel.change_set_name)

        mock_describe_change_set.return_value = {
            "Status": 'UNKOWN_STATUS',
            "ExecutionStatus": 'UNKOWN_STATUS',
        }
        with pytest.raises(UnknownStackChangeSetStatusError):
            self.stack._get_cs_status(sentinel.change_set_name)

    @patch("awsscripter.stack.Stack.describe_change_set")
    def test_get_cs_status_raises_unexpected_exceptions(
        self, mock_describe_change_set
    ):
        mock_describe_change_set.side_effect = ClientError(
            {
                "Error": {
                    "Code": "ChangeSetNotFound",
                    "Message": "ChangeSet [*] does not exist"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError):
            self.stack._get_cs_status(sentinel.change_set_name)



import click

from awsscripter.stack.helpers import catch_exceptions, confirmation
from awsscripter.stack.helpers import get_stack_or_env
from awsscripter.stack.stack_status import StackStatus


@click.command(name="create")
@click.argument("path")
@click.argument("change-set-name", required=False)
@click.option(
    "-y", "--yes", is_flag=True, help="Assume yes to all questions."
)
@click.pass_context
@catch_exceptions
def create_command(ctx, path, change_set_name, yes):
    """
    Creates a stack or a change set.

    Creates a stack for a given config PATH. Or if CHANGE_SET_NAME is specified
    creates a change set for stack in PATH.
    """
    action = "create"

    stack, _ = get_stack_or_env(ctx, path)
    if change_set_name:
        confirmation(action, yes, change_set=change_set_name, stack=path)
        stack.create_change_set(change_set_name)
    else:
        confirmation(action, yes, stack=path)
        response = stack.create()
        if response != StackStatus.COMPLETE:
            exit(1)



import click

from awsscripter.stack.helpers import catch_exceptions, get_stack_or_env, write


@click.command(name="status")
@click.argument("path")
@click.pass_context
@catch_exceptions
def status_command(ctx, path):
    """
    Print status of stack or environment.

    Prints the stack status or the status of the stacks within a environment
    for a given config PATH.
    """
    output_format = ctx.obj["output_format"]
    no_colour = ctx.obj["no_colour"]

    stack, env = get_stack_or_env(ctx, path)

    if stack:
        write(stack.get_status(), no_colour=no_colour)
    elif env:
        write(env.describe(), output_format=output_format, no_colour=no_colour)



import click

from awsscripter.stack.helpers import catch_exceptions, get_stack_or_env


@click.command(name="set-policy")
@click.argument("path")
@click.argument("policy-file", required=False)
@click.option(
    "-b", "--built-in", type=click.Choice(["deny-all", "allow-all"]),
    help="Specify a built in stack policy."
)
@click.pass_context
@catch_exceptions
def set_policy_command(ctx, path, policy_file, built_in):
    """
    Sets stack policy.

    Sets a specific stack policy for either a file or using a built-in policy.
    """
    stack, _ = get_stack_or_env(ctx, path)

    if built_in == 'deny-all':
        stack.lock()
    elif built_in == 'allow-all':
        stack.unlock()
    else:
        stack.set_policy(policy_file)



import click

from awsscripter.stack.helpers import catch_exceptions, get_stack_or_env
from awsscripter.stack.helpers import confirmation
from awsscripter.stack.stack_status import StackStatus


@click.command(name="delete")
@click.argument("path")
@click.argument("change-set-name", required=False)
@click.option(
    "-y", "--yes", is_flag=True, help="Assume yes to all questions."
)
@click.pass_context
@catch_exceptions
def delete_command(ctx, path, change_set_name, yes):
    """
    Deletes a stack or a change set.

    Deletes a stack for a given config PATH. Or if CHANGE_SET_NAME is specified
    deletes a change set for stack in PATH.
    """
    action = "delete"

    stack, env = get_stack_or_env(ctx, path)

    if stack:
        if change_set_name:
            confirmation(action, yes, change_set=change_set_name, stack=path)
            stack.delete_change_set(change_set_name)
        else:
            confirmation(action, yes, stack=path)
            response = stack.delete()
            if response != StackStatus.COMPLETE:
                exit(1)
    elif env:
        confirmation(action, yes, environment=path)
        response = env.delete()
        if not all(
            status == StackStatus.COMPLETE for status in response.values()
        ):
            exit(1)



import click
from awsscripter.stack.helpers import catch_exceptions, confirmation
from awsscripter.pciaudit.PciAuditor import PciAuditor
import logging
@click.command(name="PCIDSS")
@click.option(
    "--level", type=click.Choice(["FULL", "LESS"]), default="FULL",
    help="The level of audit , default is full")
@click.pass_context
@catch_exceptions
def audit_pcidss_command(ctx, level):
    """
    Depending on level , it will perform the CISP audit, by default the level is FULL
    """
    logger = logging.getLogger(__name__)
    logger.info("Auditing with level  " + level)
    auditor = PciAuditor("myname", "myproject", "us-east-1")
    auditor.handle("test", "test")


import datetime
import errno
import logging
import os
from uuid import UUID

import click
import pytest
import yaml
from botocore.exceptions import ClientError
from click.testing import CliRunner
from mock import MagicMock, patch, sentinel

from awsscripter.cli import cli
from awsscripter.common.config_reader import ConfigReader
from awsscripter.common.environment import Environment
from awsscripter.common.exceptions import awsscripterException
from awsscripter.stack.stack import Stack
from awsscripter.stack.helpers import CustomJsonEncoder, catch_exceptions
from awsscripter.stack.helpers import get_stack_or_env
from awsscripter.stack.helpers import setup_logging, write, ColouredFormatter
from awsscripter.stack.stack_status import StackStatus, StackChangeSetStatus


class TestCli(object):

    def setup_method(self, test_method):
        self.patcher_ConfigReader = patch("awsscripter.common.config_reader.ConfigReader")
        self.patcher_getcwd = patch("awsscripter.cli.os.getcwd")

        self.mock_ConfigReader = self.patcher_ConfigReader.start()
        self.mock_getcwd = self.patcher_getcwd.start()

        self.mock_config_reader = MagicMock(spec=ConfigReader)
        self.mock_stack = MagicMock(spec=Stack)
        self.mock_environment = MagicMock(spec=Environment)

        self.mock_config_reader.construct_stack.return_value = self.mock_stack
        self.mock_config_reader.construct_environment.return_value = \
            self.mock_environment

        self.mock_ConfigReader.return_value = self.mock_config_reader
        self.mock_getcwd.return_value = sentinel.cwd

        self.runner = CliRunner()

    def teardown_method(self, test_method):
        self.patcher_ConfigReader.stop()
        self.patcher_getcwd.stop()

    @patch("sys.exit")
    def test_catch_excecptions(self, mock_exit):
        @catch_exceptions
        def raises_exception():
            raise awsscripterException()

        raises_exception()
        mock_exit.assert_called_once_with(1)

    @pytest.mark.parametrize("command,files,output", [
        # one --var option
        (
            ["--var", "a=1", "noop"],
            {},
            {"a": "1"}
        ),
        # multiple --var options
        (
            ["--var", "a=1", "--var", "b=2", "noop"],
            {},
            {"a": "1", "b": "2"}
        ),
        # one --var-file option
        (
            ["--var-file", "foo.yaml", "noop"],
            {
                "foo.yaml": {"key1": "val1", "key2": "val2"}
            },
            {"key1": "val1", "key2": "val2"}
        ),
        # multiple --var-file option
        (
            ["--var-file", "foo.yaml", "--var-file", "bar.yaml", "noop"],
            {
                "foo.yaml": {"key1": "parent_value1", "key2": "parent_value2"},
                "bar.yaml": {"key2": "child_value2", "key3": "child_value3"}
            },
            {
                "key1": "parent_value1",
                "key2": "child_value2",
                "key3": "child_value3"
            }
        ),
        # mix of --var and --var-file
        (
            ["--var-file", "foo.yaml", "--var", "key2=var2", "noop"],
            {
                "foo.yaml": {"key1": "file1", "key2": "file2"}
            },
            {"key1": "file1", "key2": "var2"}
        ),
    ])
    def test_user_variables(self, command, files, output):
        @cli.command()
        @click.pass_context
        def noop(ctx):
            click.echo(yaml.safe_dump(ctx.obj["user_variables"]))

        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            for name, content in files.items():
                with open(name, "w") as fh:
                    yaml.safe_dump(content, fh)

            result = self.runner.invoke(cli, command)
        self.patcher_getcwd.start()

        user_variables = yaml.safe_load(result.output)
        assert result.exit_code == 0
        assert user_variables == output

    def test_validate_template_with_valid_template(self):
        self.mock_stack.template.validate.return_value = {
                    "Parameters": "Example",
                    "ResponseMetadata": {
                        "HTTPStatusCode": 200
                    }
                }
        result = self.runner.invoke(cli, ["validate", "dev/vpc.yaml"])
        self.mock_stack.template.validate.assert_called_with()

        assert result.output == "Template is valid. Template details:\n\n" \
            "Parameters: Example\n\n"

    def test_validate_template_with_invalid_template(self):
        client_error = ClientError(
            {
                "Errors":
                {
                    "Message": "Unrecognized resource types",
                    "Code": "ValidationError",
                }
            },
            "ValidateTemplate"
        )
        self.mock_stack.template.validate.side_effect = client_error

        expected_result = str(client_error) + "\n"
        result = self.runner.invoke(cli, ["validate", "dev/vpc.yaml"])
        assert result.output == expected_result

    def test_generate_template(self):
        self.mock_stack.template.body = "body"
        result = self.runner.invoke(cli, ["generate", "dev/vpc.yaml"])
        self.mock_config_reader.construct_stack.assert_called_with(
            "dev/vpc.yaml"
        )

        assert result.output == "body\n"

    def test_lock_stack(self):
        self.runner.invoke(
            cli, ["set-policy", "dev/vpc.yaml", "-b", "deny-all"]
        )
        self.mock_config_reader.construct_stack.assert_called_with(
                "dev/vpc.yaml"
        )
        self.mock_stack.lock.assert_called_with()

    def test_unlock_stack(self):
        self.runner.invoke(
            cli, ["set-policy", "dev/vpc.yaml", "-b", "allow-all"]
        )
        self.mock_config_reader.construct_stack.assert_called_with(
                "dev/vpc.yaml"
        )
        self.mock_stack.unlock.assert_called_with()

    def test_set_policy_with_file_flag(self):
        policy_file = "tests/fixtures/stack_policies/lock.json"
        result = self.runner.invoke(cli, [
            "set-policy", "dev/vpc.yaml", policy_file
        ])
        assert result.exit_code == 0
        self.mock_stack.set_policy.assert_called_once_with(policy_file)

    def test_describe_policy_with_existing_policy(self):
        self.mock_stack.get_policy.return_value = {
            "StackPolicyBody": "Body"
        }

        result = self.runner.invoke(
            cli, ["describe", "policy", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert result.output == "Body\n"

    def test_list_env_resources(self):
        response = {
            "stack-name-1": {
                "StackResources": [
                    {
                        "LogicalResourceId": "logical-resource-id",
                        "PhysicalResourceId": "physical-resource-id"
                    }
                ]
            },
            "stack-name-2": {
                "StackResources": [
                    {
                        "LogicalResourceId": "logical-resource-id",
                        "PhysicalResourceId": "physical-resource-id"
                    }
                ]
            }
        }
        self.mock_environment.describe_resources.return_value = response
        result = self.runner.invoke(cli, ["list", "resources", "dev"])
        assert yaml.load(result.output) == response
        assert result.exit_code == 0

    def test_list_stack_resources(self):
        response = {
                "StackResources": [
                    {
                        "LogicalResourceId": "logical-resource-id",
                        "PhysicalResourceId": "physical-resource-id"
                    }
                ]
            }
        self.mock_stack.describe_resources.return_value = response
        result = self.runner.invoke(cli, ["stack", "list", "resources", "dev/vpc.yaml"])
        assert yaml.load(result.output) == response
        assert result.exit_code == 0

    def test_create_envorstack(self):
        result = self.runner.invoke(cli, ["--debug","--dir","D:\\mycode\\examples\\baseecscluster","stack", "launch", "dev"])
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "command,success,yes_flag,exit_code", [
            ("create", True, True, 0),
            ("create", False, True, 1),
            ("create", True, False, 0),
            ("create", False, False, 1),
            ("delete", True, True, 0),
            ("delete", False, True, 1),
            ("delete", True, False, 0),
            ("delete", False, False, 1),
            ("update", True, True, 0),
            ("update", False, True, 1),
            ("update", True, False, 0),
            ("update", False, False, 1),
            ("launch", True, True, 0),
            ("launch", False, True, 1),
            ("launch", True, False, 0),
            ("launch", False, False, 1)
        ]
    )
    def test_stack_commands(self, command, success, yes_flag, exit_code):
        getattr(self.mock_stack, command).return_value = \
            StackStatus.COMPLETE if success else StackStatus.FAILED

        kwargs = {"args": [command, "dev/vpc.yaml"]}
        if yes_flag:
            kwargs["args"].append("-y")
        else:
            kwargs["input"] = "y\n"

        result = self.runner.invoke(cli, **kwargs)

        getattr(self.mock_stack, command).assert_called_with()
        assert result.exit_code == exit_code

    @pytest.mark.parametrize(
        "command,yes_flag", [
            ("create", True),
            ("create", False),
            ("delete", True),
            ("delete", False),
            ("execute", True),
            ("execute", False)
        ]
    )
    def test_change_set_commands(self, command, yes_flag):
        stack_command = command + "_change_set"

        kwargs = {"args": [command, "dev/vpc.yaml", "cs1"]}
        if yes_flag:
            kwargs["args"].append("-y")
        else:
            kwargs["input"] = "y\n"

        result = self.runner.invoke(cli, **kwargs)

        getattr(self.mock_stack, stack_command).assert_called_with("cs1")
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "command,success,yes_flag,exit_code", [
            ("delete", True, True, 0),
            ("delete", False, True, 1),
            ("delete", True, False, 0),
            ("delete", False, False, 1),
            ("launch", True, True, 0),
            ("launch", False, True, 1),
            ("launch", True, False, 0),
            ("launch", False, False, 1)
        ]
    )
    def test_env_commands(self, command, success, yes_flag, exit_code):
        status = StackStatus.COMPLETE if success else StackStatus.FAILED
        response = {"stack": status}

        getattr(self.mock_environment, command).return_value = response

        kwargs = {"args": [command, "dev"]}
        if yes_flag:
            kwargs["args"].append("-y")
        else:
            kwargs["input"] = "y\n"

        result = self.runner.invoke(cli, **kwargs)

        getattr(self.mock_environment, command).assert_called_with()
        assert result.exit_code == exit_code

    @patch('awsscripter.cli.update.uuid1')
    @patch('awsscripter.cli.update.write')
    @pytest.mark.parametrize(
        "verbose_flag,yes_flag,cs_success", [
            (False, True, True),
            (False, False, True),
            (True, True, True),
            (True, False, True),
            (False, True, False),
            (False, False, False),
            (True, True, False),
            (True, False, False)
        ]
    )
    def test_update_with_change_set_with_input_yes(
        self, mock_write, mock_uuid1, verbose_flag, yes_flag, cs_success
    ):
        self.mock_stack.wait_for_cs_completion.return_value = \
            StackChangeSetStatus.READY if cs_success \
            else StackChangeSetStatus.DEFUNCT
        response = {
            "VerboseProperty": "VerboseProperty",
            "ChangeSetName": "ChangeSetName",
            "CreationTime": "CreationTime",
            "ExecutionStatus": "ExecutionStatus",
            "StackName": "StackName",
            "Status": "Status",
            "StatusReason": "StatusReason",
            "Changes": [
                {
                    "ResourceChange": {
                        "Action": "Action",
                        "LogicalResourceId": "LogicalResourceId",
                        "PhysicalResourceId": "PhysicalResourceId",
                        "Replacement": "Replacement",
                        "ResourceType": "ResourceType",
                        "Scope": "Scope",
                        "VerboseProperty": "VerboseProperty"
                    }
                }
            ]
        }
        self.mock_stack.describe_change_set.return_value = response

        mock_uuid1.return_value = UUID(int=0)

        kwargs = {"args": ["update", "dev/vpc.yaml", "-c"]}

        if verbose_flag:
            kwargs["args"].append("-v")

        if yes_flag:
            kwargs["args"].append("-y")
        else:
            kwargs["input"] = "y\n"

        result = self.runner.invoke(cli, **kwargs)

        if cs_success:
            if not verbose_flag:
                del response["VerboseProperty"]
                del response["Changes"][0]["ResourceChange"]["VerboseProperty"]
            mock_write.assert_called_once_with(response, 'yaml')
            self.mock_stack.execute_change_set.assert_called_once_with(
                "change-set-00000000000000000000000000000000"
            )

        self.mock_stack.delete_change_set.assert_called_once_with(
            "change-set-00000000000000000000000000000000"
        )

        assert result.exit_code == (0 if cs_success else 1)

    @pytest.mark.parametrize(
        "verbose_flag,", [
            (False),
            (True)
        ]
    )
    def test_describe_change_set(self, verbose_flag):
        response = {
            "VerboseProperty": "VerboseProperty",
            "ChangeSetName": "ChangeSetName",
            "CreationTime": "CreationTime",
            "ExecutionStatus": "ExecutionStatus",
            "StackName": "StackName",
            "Status": "Status",
            "StatusReason": "StatusReason",
            "Changes": [
                {
                    "ResourceChange": {
                        "Action": "Action",
                        "LogicalResourceId": "LogicalResourceId",
                        "PhysicalResourceId": "PhysicalResourceId",
                        "Replacement": "Replacement",
                        "ResourceType": "ResourceType",
                        "Scope": "Scope",
                        "VerboseProperty": "VerboseProperty"
                    }
                }
            ]
        }
        args = ["describe", "change-set", "dev/vpc.yaml", "cs1"]
        if verbose_flag:
            args.append("-v")

        self.mock_stack.describe_change_set.return_value = response
        result = self.runner.invoke(cli, args)
        if not verbose_flag:
            del response["VerboseProperty"]
            del response["Changes"][0]["ResourceChange"]["VerboseProperty"]
        assert yaml.load(result.output) == response
        assert result.exit_code == 0

    def test_list_change_sets_with_200(self):
        self.mock_stack.list_change_sets.return_value = {
            "ResponseMetadata": {
                "HTTPStatusCode": 200
            },
            "ChangeSets": "Test"
        }
        result = self.runner.invoke(
            cli, ["list", "change-sets", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert yaml.load(result.output) == {"ChangeSets": "Test"}

    def test_list_change_sets_without_200(self):
        response = {
            "ResponseMetadata": {
                "HTTPStatusCode": 404
            },
            "ChangeSets": "Test"
        }
        self.mock_stack.list_change_sets.return_value = response

        result = self.runner.invoke(
            cli, ["list", "change-sets", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert yaml.load(result.output) == response

    def test_list_outputs(self):
        outputs = [{"OutputKey": "Key", "OutputValue": "Value"}]
        self.mock_stack.describe_outputs.return_value = outputs
        result = self.runner.invoke(
            cli, ["list", "outputs", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert yaml.load(result.output) == outputs

    def test_list_outputs_with_export(self):
        outputs = [{"OutputKey": "Key", "OutputValue": "Value"}]
        self.mock_stack.describe_outputs.return_value = outputs
        result = self.runner.invoke(
            cli, ["list", "outputs", "dev/vpc.yaml", "-e", "envvar"]
        )
        assert result.exit_code == 0
        assert yaml.load(result.output) == "export SCEPTRE_Key=Value"

    def test_status_with_env(self):
        self.mock_environment.describe.return_value = {"stack": "status"}

        result = self.runner.invoke(cli, ["status", "dev"])
        assert result.exit_code == 0
        assert result.output == "stack: status\n\n"

    def test_status_with_stack(self):
        self.mock_stack.get_status.return_value = "status"

        result = self.runner.invoke(cli, ["status", "dev/vpc.yaml"])
        assert result.exit_code == 0
        assert result.output == "status\n"

    def test_init_project_non_existant(self):
        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            sceptre_dir = os.path.abspath('./example')
            config_dir = os.path.join(sceptre_dir, "config")
            template_dir = os.path.join(sceptre_dir, "templates")
            region = "test-region"
            os.environ["AWS_DEFAULT_REGION"] = region
            defaults = {
                "project_code": "example",
                "region": region
            }

            result = self.runner.invoke(cli, ["init1", "project", "example"])
            assert not result.exception
            assert os.path.isdir(config_dir)
            assert os.path.isdir(template_dir)

            with open(os.path.join(config_dir, "config.yaml")) as config_file:
                config = yaml.load(config_file)

            assert config == defaults
        self.patcher_getcwd.start()

    def test_init_project_already_exist(self):
        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            sceptre_dir = os.path.abspath('./example')
            config_dir = os.path.join(sceptre_dir, "config")
            template_dir = os.path.join(sceptre_dir, "templates")
            existing_config = {"Test": "Test"}

            os.mkdir(sceptre_dir)
            os.mkdir(config_dir)
            os.mkdir(template_dir)

            config_filepath = os.path.join(config_dir, "config.yaml")
            with open(config_filepath, 'w') as config_file:
                yaml.dump(existing_config, config_file)

            result = self.runner.invoke(cli, ["init1", "project", "example"])
            assert result.exit_code == 1
            assert result.output == 'Folder \"example\" already exists.\n'
            assert os.path.isdir(config_dir)
            assert os.path.isdir(template_dir)

            with open(os.path.join(config_dir, "config.yaml")) as config_file:
                config = yaml.load(config_file)
            assert existing_config == config
        self.patcher_getcwd.start()

    def test_init_project_another_exception(self):
        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            patcher_mkdir = patch("awsscripter.cli.init1.os.mkdir")
            mock_mkdir = patcher_mkdir.start()
            mock_mkdir.side_effect = OSError(errno.EINVAL)
            result = self.runner.invoke(cli, ["init1", "project", "example"])
            mock_mkdir = patcher_mkdir.stop()
            assert str(result.exception) == str(OSError(errno.EINVAL))
        self.patcher_getcwd.start()

    @pytest.mark.parametrize("environment,config_structure,stdin,result", [
        (
         "A",
         {"": {}},
         'y\nA\nA\n', {"project_code": "A", "region": "A"}
        ),
        (
         "A",
         {"": {"project_code": "top", "region": "top"}},
         'y\n\n\n', {}
        ),
        (
         "A",
         {"": {"project_code": "top", "region": "top"}},
         'y\nA\nA\n', {"project_code": "A", "region": "A"}
        ),
        (
         "A/A",
         {
            "": {"project_code": "top", "region": "top"},
            "A": {"project_code": "A", "region": "A"},
         },
         'y\nA/A\nA/A\n', {"project_code": "A/A", "region": "A/A"}
        ),
        (
         "A/A",
         {
            "": {"project_code": "top", "region": "top"},
            "A": {"project_code": "A", "region": "A"},
         },
         'y\nA\nA\n', {}
        )
    ])
    def test_init_environment(
        self, environment, config_structure, stdin, result
    ):
        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            sceptre_dir = os.path.abspath('./example')
            config_dir = os.path.join(sceptre_dir, "config")
            os.makedirs(config_dir)

            env_dir = os.path.join(sceptre_dir, "config", environment)
            for env_path, config in config_structure.items():
                path = os.path.join(config_dir, env_path)
                try:
                    os.makedirs(path)
                except OSError as e:
                    if e.errno == errno.EEXIST and os.path.isdir(path):
                        pass
                    else:
                        raise

                filepath = os.path.join(path, "config.yaml")
                with open(filepath, 'w') as config_file:
                    yaml.safe_dump(
                        config, stream=config_file, default_flow_style=False
                    )

            os.chdir(sceptre_dir)

            cmd_result = self.runner.invoke(
                cli, ["init1", "env", environment],
                input=stdin
            )

            if result:
                with open(os.path.join(env_dir, "config.yaml")) as config_file:
                    config = yaml.load(config_file)
                assert config == result
            else:
                assert cmd_result.output.endswith(
                    "No config.yaml file needed - covered by parent config.\n"
                )
        self.patcher_getcwd.start()

    def test_init_environment_with_existing_folder(self):
        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            sceptre_dir = os.path.abspath('./example')
            config_dir = os.path.join(sceptre_dir, "config")
            env_dir = os.path.join(config_dir, "A")

            os.makedirs(env_dir)
            os.chdir(sceptre_dir)

            cmd_result = self.runner.invoke(
                cli, ["init1", "env", "A"], input="y\n\n\n"
            )

            assert cmd_result.output.startswith(
                "Environment path exists. "
                "Do you want initialise config.yaml?"
            )
            with open(os.path.join(env_dir, "config.yaml")) as config_file:
                config = yaml.load(config_file)
            assert config == {"project_code": "", "region": ""}

        self.patcher_getcwd.start()

    def test_init_environment_with_another_exception(self):
        self.patcher_getcwd.stop()
        with self.runner.isolated_filesystem():
            sceptre_dir = os.path.abspath('./example')
            config_dir = os.path.join(sceptre_dir, "config")
            env_dir = os.path.join(config_dir, "A")

            os.makedirs(env_dir)
            os.chdir(sceptre_dir)

            patcher_mkdir = patch("awsscripter.cli.init1.os.mkdir")
            mock_mkdir = patcher_mkdir.start()
            mock_mkdir.side_effect = OSError(errno.EINVAL)
            result = self.runner.invoke(cli, ["init1", "env", "A"])
            mock_mkdir = patcher_mkdir.stop()
            assert str(result.exception) == str(OSError(errno.EINVAL))

        self.patcher_getcwd.start()

    def test_setup_logging_with_debug(self):
        logger = setup_logging(True, False)
        assert logger.getEffectiveLevel() == logging.DEBUG
        assert logging.getLogger("botocore").getEffectiveLevel() == \
            logging.INFO

        # Silence logging for the rest of the tests
        logger.setLevel(logging.CRITICAL)

    def test_setup_logging_without_debug(self):
        logger = setup_logging(False, False)
        assert logger.getEffectiveLevel() == logging.INFO
        assert logging.getLogger("botocore").getEffectiveLevel() == \
            logging.CRITICAL

        # Silence logging for the rest of the tests
        logger.setLevel(logging.CRITICAL)

    @patch("awsscripter.cli.click.echo")
    def test_write_with_yaml_format(self, mock_echo):
        write({"key": "value"}, "yaml")
        mock_echo.assert_called_once_with("key: value\n")

    @patch("awsscripter.cli.click.echo")
    def test_write_with_json_format(self, mock_echo):
        write({"key": "value"}, "json")
        mock_echo.assert_called_once_with('{"key": "value"}')

    @patch("awsscripter.cli.click.echo")
    def test_write_status_with_colour(self, mock_echo):
        write("stack: CREATE_COMPLETE", no_colour=False)
        mock_echo.assert_called_once_with(
            "stack: \x1b[32mCREATE_COMPLETE\x1b[0m"
        )

    @patch("awsscripter.cli.click.echo")
    def test_write_status_without_colour(self, mock_echo):
        write("stack: CREATE_COMPLETE", no_colour=True)
        mock_echo.assert_called_once_with("stack: CREATE_COMPLETE")

    @patch("awsscripter.cli.helpers.StackStatusColourer.colour")
    @patch("awsscripter.cli.helpers.logging.Formatter.format")
    def test_ColouredFormatter_format_with_string(
            self, mock_format, mock_colour
    ):
        mock_format.return_value = sentinel.response
        mock_colour.return_value = sentinel.coloured_response
        coloured_formatter = ColouredFormatter()
        response = coloured_formatter.format("string")
        mock_format.assert_called_once_with("string")
        mock_colour.assert_called_once_with(sentinel.response)
        assert response == sentinel.coloured_response

    def test_CustomJsonEncoder_with_non_json_serialisable_object(self):
        encoder = CustomJsonEncoder()
        response = encoder.encode(datetime.datetime(2016, 5, 3))
        assert response == '"2016-05-03 00:00:00"'

    def test_get_stack_or_env_with_stack(self):
        ctx = MagicMock(obj={
            "sceptre_dir": sentinel.sceptre_dir,
            "user_variables": sentinel.user_variables
        })
        stack, env = get_stack_or_env(ctx, "stack.yaml")
        self.mock_ConfigReader.assert_called_once_with(
            sentinel.sceptre_dir, sentinel.user_variables
        )
        assert isinstance(stack, Stack)
        assert env is None

    def test_get_stack_or_env_with_nested_stack(self):
        ctx = MagicMock(obj={
            "sceptre_dir": sentinel.sceptre_dir,
            "user_variables": sentinel.user_variables
        })
        stack, env = get_stack_or_env(ctx, "environment/dir/stack.yaml")
        self.mock_ConfigReader.assert_called_once_with(
            sentinel.sceptre_dir, sentinel.user_variables
        )
        assert isinstance(stack, Stack)
        assert env is None

    def test_get_stack_or_env_with_env(self):
        ctx = MagicMock(obj={
            "sceptre_dir": sentinel.sceptre_dir,
            "user_variables": sentinel.user_variables
        })
        stack, env = get_stack_or_env(ctx, "environment/dir")
        self.mock_ConfigReader.assert_called_once_with(
            sentinel.sceptre_dir, sentinel.user_variables
        )
        assert isinstance(env, Environment)
        assert stack is None
def main():
    foo = TestCli()  
    # foo.setup_method("test_list_stack_resources")
    # foo.test_list_stack_resources()
    foo.setup_method("test_create_envorstack")
    foo.test_create_envorstack()


if __name__ == "__main__":
    main()


"""Base class for implementing AWS functionalities as classes.
Add additional features here common to all your AWS functionalities , like querying status."""


class AwsBase(object):
    def get_status(self):
        raise NotImplementedError


import click
from awsscripter.cli.testcommand.command1 import testcommand_command1
from awsscripter.cli.testcommand.command2 import testcommand_command2

@click.group(name="testcommand")
def testcommand_group():
    """
    A sample testcommand

    """
    pass
testcommand_group.add_command(testcommand_command1)
testcommand_group.add_command(testcommand_command2)


import click

from awsscripter.stack.helpers import catch_exceptions, get_stack_or_env, write


@click.group(name="list")
def list_group():
    """
    Commands for listing attributes of stacks.

    """
    pass


@list_group.command(name="resources")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_resources(ctx, path):
    """
    List resources for stack or environment.

    """
    stack, env = get_stack_or_env(ctx, path)
    output_format = ctx.obj["output_format"]

    if stack:
        write(stack.describe_resources(), output_format)
    elif env:
        write(env.describe_resources(), output_format)


@list_group.command(name="outputs")
@click.argument("path")
@click.option(
    "-e", "--export", type=click.Choice(["envvar"]),
    help="Specify the export formatting."
)
@click.pass_context
@catch_exceptions
def list_outputs(ctx, path, export):
    """
    List outputs for stack.

    """
    stack, _ = get_stack_or_env(ctx, path)
    response = stack.describe_outputs()

    if export == "envvar":
        write("\n".join(
            [
                "export awsscripter_{0}={1}".format(
                    output["OutputKey"], output["OutputValue"]
                )
                for output in response
            ]
        ))
    else:
        write(response, ctx.obj["output_format"])


@list_group.command(name="change-sets")
@click.argument("path")
@click.pass_context
@catch_exceptions
def list_change_sets(ctx, path):
    """
    List change sets for stack.

    """
    stack, _ = get_stack_or_env(ctx, path)
    response = stack.list_change_sets()
    if response['ResponseMetadata']['HTTPStatusCode'] == 200:
        del response['ResponseMetadata']
    write(response, ctx.obj["output_format"])



# -*- coding: utf-8 -*-

"""
awsscripter.config

This module implements a Config class, which stores a stack or environment's
configuration.
"""

import collections
import copy
import datetime
import logging
from functools import partial
from glob import glob
from os import path, environ

import jinja2
import yaml
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from pkg_resources import iter_entry_points

from awsscripter import __version__
from awsscripter.common.environment import Environment
from awsscripter.common.exceptions import EnvironmentNotFoundError, VersionIncompatibleError
from awsscripter.common.exceptions import InvalidawsscripterDirectoryError, ConfigFileNotFoundError
from awsscripter.stack.stack import Stack

ConfigAttributes = collections.namedtuple("Attributes", "required optional")

ENVIRONMENT_CONFIG_ATTRIBUTES = ConfigAttributes(
    {
        "project_code",
        "region"
    },
    {
        "iam_role",
        "template_bucket_name",
        "template_key_prefix",
        "require_version"
    }
)

STACK_CONFIG_ATTRIBUTES = ConfigAttributes(
    {
        "template_path"
    },
    {
        "dependencies",
        "hooks",
        "parameters",
        "protect",
        "awsscripter_user_data",
        "stack_name",
        "stack_tags",
        "role_arn"
    }
)


class ConfigReader(object):
    """
    Respresents a awsscripter project folder. Reads in yaml configuration files and
    produces Stack or Environment objects. Responsible for loading Resolvers
    and Hook classes and adding them as constructors to the PyYAML parser.

    :param awsscripter_dir: The absolute path to the awsscripter directory.
    :type awsscripter_dir: str
    """

    def __init__(self, awsscripter_dir, variables=None):
        self.logger = logging.getLogger(__name__)

        self.awsscripter_dir = awsscripter_dir

        # Check is valid awsscripter project folder
        self.config_folder = path.join(self.awsscripter_dir, "config")
        self._check_valid_awsscripter_dir(self.config_folder)

        # Add Resolver and Hook classes to PyYAML loader
        self._add_yaml_constructors(
            ["awsscripter.hooks", "awsscripter.resolvers"]
        )

        if not variables:
            variables = {}

        self.templating_vars = {"var": variables}

    def _add_yaml_constructors(self, entry_point_groups):
        """
        Adds PyYAML constructor functions for all classes found registered at
        the given entry point groups. Classes are registered whereby the node
        tag is the entry point name.

        :param entry_point_groups: Names of entry point groups.
        :type entry_point_groups: list
        """
        self.logger.debug(
            "Adding yaml constructors for the entry point groups {0}".format(
                entry_point_groups
            )
        )

        def constructor_factory(node_class):
            """
            Returns a partial function that will initialise objects from a
            given node class.

            :param node_class: Class representing the node.
            :type node_class: class
            :returns: Partial function of class initialiser.
            :rtype: partial
            """
            # This function signture is required by PyYAML
            def class_constructor(loader, node):
                # Returning partial as initialisation needs to be deferred
                # until config file has been completely parsed. This because
                # the config itself is passed to the node_class and can be used
                # during initialisation. Construct_nodes function will call any
                # partials within the config to finially initialise objects
                # after the config has been loaded by PyYAML.
                return partial(
                    node_class, loader.construct_scalar(node)
                )  # pragma: no cover

            return class_constructor

        for group in entry_point_groups:
            for entry_point in iter_entry_points(group):
                # Retrieve name and class from entry point
                node_tag = u'!' + entry_point.name
                node_class = entry_point.load()

                # Add constructor to PyYAML loader
                yaml.SafeLoader.add_constructor(
                    node_tag, constructor_factory(node_class)
                )
                self.logger.debug(
                    "Added constructor for %s with node tag %s",
                    str(node_class), node_tag
                )

    def read(self, rel_path, base_config=None):
        """
        Reads in configuration from yaml files within the awsscripter project
        folder.

        :param rel_path: Relative path to config to read.
        :type rel_path: str
        :param base_config: Base config to provide defaults.
        :type base_config: dict
        :returns: Config read from config files.
        :rtype: dict
        """
        self.logger.debug("Reading in '%s' files...", rel_path)
        directory_path, filename = path.split(rel_path)
        abs_path = path.join(self.config_folder, rel_path)

        # Adding properties from class
        config = {
            "awsscripter_dir": self.awsscripter_dir,
            "environment_path": directory_path
        }

        # Adding defaults from base config.
        if base_config:
            config.update(base_config)

        # Check is file exists, but ignore config.yaml as can be inherited.
        if not path.isfile(abs_path) and not filename.endswith("config.yaml"):
            raise ConfigFileNotFoundError(
                "Config file \"{0}\" not found.".format(rel_path)
            )

        # Parse and read in the config files.
        config.update(self._recursive_read(directory_path, filename))
        self._check_version(config)

        self.logger.debug("Config: %s", config)
        return config

    def _recursive_read(self, directory_path, filename):
        """
        Traverses the directory_path, from top to bottom, reading in all
        relevant config files. If config items appear in files lower down the
        environment tree, they overwrite items from further up.

        :param directory_path: Relative directory path to config to read.
        :type directory_path: str
        :param filename: Base config to provide defaults.
        :type filename: dict
        :returns: Representation of inherited config.
        :rtype: dict
        """

        parent_directory = path.split(directory_path)[0]

        # Base condition for recursion
        config = {}

        if directory_path:
            config = self._recursive_read(parent_directory, filename)

        # Read config file and overwrite inherited properties
        child_config = self._read(directory_path, filename) or {}
        config.update(child_config)

        return config

    def _read(self, directory_path, basename):
        """
        Traverses the directory_path, from top to bottom, reading in all
        relevant config files. If config items appear in files lower down the
        environment tree, they overwrite items from further up.

        :param directory_path: Relative directory path to config to read.
        :type directory_path: str
        :param filename: Base config to provide defaults.
        :type filename: dict
        :returns: Representation of inherited config.
        :rtype: dict
        """
        abs_directory_path = path.join(self.config_folder, directory_path)
        if path.isfile(path.join(abs_directory_path, basename)):
            env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(abs_directory_path),
                undefined=jinja2.StrictUndefined
            )
            template = env.get_template(basename)
            rendered_template = template.render(
                environment_variable=environ,
                environment_path=directory_path.split("/"),
                **self.templating_vars
            )
            config = yaml.safe_load(rendered_template)
            return config

    @staticmethod
    def _check_valid_awsscripter_dir(config_path):
        """
        Raises an InvalidawsscripterDirectoryError if ``path`` is not a directory.

        :param path: A config directory path.
        :type path: str
        :raises: awsscripter.common.exceptions.InvalidawsscripterDirectoryError
        """
        if not path.isdir(config_path):
            raise InvalidawsscripterDirectoryError(
                "Check '{0}' exists.".format(config_path)
            )

    def _check_version(self, config):
        """
        Raises a VersionIncompatibleException when the current awsscripter version
        does not comply with the configured version requirement.

        :raises: awsscripter.common.exceptions.VersionIncompatibleException
        """
        awsscripter_version = __version__
        if 'require_version' in config:
            require_version = config['require_version']
            if Version(awsscripter_version) not in SpecifierSet(require_version):
                raise VersionIncompatibleError(
                    "Current awsscripter version ({0}) does not meet version "
                    "requirements: {1}".format(
                        awsscripter_version, require_version
                    )
                )

    @classmethod
    def _construct_nodes(cls, attr, stack):
        """
        Search through a data structure to call any partial functions to
        finialise contruction of YAML objects.

        :param attr: Data structure to search through.
        :type attr: dict
        :param stack: Stack object to associate to objects.
        :type directory_path: awsscripter.stack.stack.Stack
        """
        if isinstance(attr, dict):
            for key, value in attr.items():
                if isinstance(value, partial):
                    attr[key] = value(stack)
                elif isinstance(value, list) or isinstance(value, dict):
                    cls._construct_nodes(value, stack)
        elif isinstance(attr, list):
            for index, value in enumerate(attr):
                if isinstance(value, partial):
                    attr[index] = value(stack)
                elif isinstance(value, list) or isinstance(value, dict):
                    cls._construct_nodes(value, stack)

    @staticmethod
    def _collect_s3_details(stack_name, config):
        """
        Collects and constructs details for where to store the template in S3.

        :param stack_name: Stack name.
        :type stack_name: str
        :param config: Config with details.
        :type config: dict
        :returns: S3 details.
        :rtype: dict
        """
        s3_details = None
        if "template_bucket_name" in config:
            template_key = "/".join([
                stack_name, "{time_stamp}.json".format(
                    time_stamp=datetime.datetime.utcnow().strftime(
                        "%Y-%m-%d-%H-%M-%S-%fZ"
                    )
                )
            ])

            if "template_key_prefix" in config:
                prefix = config["template_key_prefix"]
                template_key = "/".join([prefix.strip("/"), template_key])

            s3_details = {
                 "bucket_name": config["template_bucket_name"],
                 "bucket_key": template_key
            }
        return s3_details

    def _construct_stack(self, rel_path, environment_config=None):
        """
        Construct a Stack object from a config path and a base config.

        :param rel_path: A relative config file path.
        :type rel_path: str
        :param config: Base config to use as defaults.
        :type config: dict
        :returns: Stack object
        :rtype: awsscripter.stack.stack.Stack
        """
        directory, filename = path.split(rel_path)
        if filename != "config.yaml":
            self.templating_vars["environment_config"] = environment_config
            config = self.read(rel_path, environment_config)
            stack_name = path.splitext(rel_path)[0]
            abs_template_path = path.join(
                self.awsscripter_dir, config["template_path"]
            )

            s3_details = self._collect_s3_details(
                stack_name, config
            )

            stack = Stack(
                name=stack_name,
                project_code=config["project_code"],
                template_path=abs_template_path,
                region=config["region"],
                iam_role=config.get("iam_role"),
                parameters=config.get("parameters", {}),
                awsscripter_user_data=config.get("awsscripter_user_data", {}),
                hooks=config.get("hooks", {}),
                s3_details=s3_details,
                dependencies=config.get("dependencies", []),
                role_arn=config.get("role_arn"),
                protected=config.get("protect", False),
                tags=config.get("stack_tags", {}),
                external_name=config.get("external_name"),
                notifications=config.get("notifications"),
                on_failure=config.get("on_failure")
            )

            self._construct_nodes(config, stack)
            del self.templating_vars["environment_config"]
            return stack

    def construct_stack(self, rel_path):
        """
        Construct a Stack object from a config path with the environment config
        as the base config.

        :param rel_path: A relative stack config path from the config folder.
        :type rel_path: str
        """
        if not path.isfile(path.join(self.config_folder, rel_path)):
            raise ConfigFileNotFoundError(
                "Config file not found for '{}'".format(rel_path)
            )
        directory = path.split(rel_path)[0]
        environment_config = self.read(path.join(directory, "config.yaml"))
        return self._construct_stack(rel_path, environment_config)

    def construct_environment(self, rel_path):
        """
        Construct an Environment object from a environment path with all
        associated sub-environments and stack objects.

        :param rel_path: A relative environment path from the config folder.
        :type rel_path: str
        """
        if not path.isdir(path.join(self.config_folder, rel_path)):
            raise EnvironmentNotFoundError(
                "Environment not found for '{}'".format(rel_path)
            )
        environment_config = self.read(path.join(rel_path, "config.yaml"))
        environment = Environment(rel_path)

        items = glob(
            path.join(self.awsscripter_dir, "config", rel_path, "*")
        )

        paths = {
            item: path.relpath(
                item, path.join(self.awsscripter_dir, "config")
            )
            for item in items if not item.endswith("config.yaml")
        }

        is_leaf = not any([path.isdir(abs_path) for abs_path in paths.keys()])

        for abs_path, rel_path in paths.items():
            if not is_leaf and path.isdir(abs_path):
                environment.sub_environments.append(
                    self.construct_environment(rel_path)
                )
            elif is_leaf and path.isfile(abs_path):
                stack = self._construct_stack(
                    rel_path, copy.deepcopy(environment_config)
                )
                environment.stacks.append(stack)

        return environment



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from awsscripter import __version__
import setuptools
from setuptools import setup,find_packages
#from distutils.core import setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

install_requirements = [
    "boto3>=1.3.0,<1.5.0",
    "click==6.6",
    "PyYaml==3.12",
    "Jinja2==2.8",
    "packaging==16.8",
    "colorama==0.3.7",
    "six==1.11.0"
]

test_requirements = [
    "pytest>=3.2",
    "troposphere>=2.0.0",
    "moto==0.4.31",
    "mock==2.0.0",
    "behave==1.2.5",
    "freezegun==0.3.9"
]

setup_requirements = [
    "pytest-runner>=3"
]

setuptools.setup(
    name="awsscripter",
    version=__version__,
    description="Cloud Provisioning Tool",
    long_description_content_type="text/markdown",
    long_description=readme,
    author="Papu",
    author_email="papu.bhattacharya@synectiks.com",
    license='Apache2',
    url="https://github.com/xformation/awsscripter",
    packages=setuptools.find_packages(),
    package_dir={
        "awsscripter": "awsscripter"
    },
    py_modules=["awsscripter"],
    entry_points={
        "console_scripts": [
            'awsscripter = awsscripter.cli:cli'
        ],
        "awsscripter.hooks": [
            "asg_scheduled_actions ="
            "awsscripter.hooks.asg_scaling_processes:ASGScalingProcesses",
            "cmd = awsscripter.hooks.cmd:Cmd"
        ],
        "awsscripter.resolvers": [
            "environment_variable ="
            "awsscripter.resolvers.environment_variable:EnvironmentVariable",
            "file_contents = awsscripter.resolvers.file_contents:FileContents",
            "stack_output = awsscripter.resolvers.stack_output:StackOutput",
            "stack_output_external ="
            "awsscripter.resolvers.stack_output:StackOutputExternal"
        ]
    },
    data_files=[
        ("awsscripter/stack_policies", [
            "awsscripter/stack_policies/lock.json",
            "awsscripter/stack_policies/unlock.json"
        ])
    ],
    include_package_data=True,
    zip_safe=False,
    keywords="awsscripter",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Environment :: Console",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6"
    ],
    test_suite="tests",
    install_requires=install_requirements,
    tests_require=test_requirements,
    setup_requires=setup_requirements,
    extras_require={
        "test": test_requirements
    }
)



# -*- coding: utf-8 -*-


class awsscripterException(Exception):
    """
    Base class for all awsscripter errors
    """
    pass


class ProjectAlreadyExistsError(awsscripterException):
    """
    Error raised when awsscripter project already exists.
    """
    pass


class InvalidawsscripterDirectoryError(awsscripterException):
    """
    Error raised if a awsscripter directory is invalid.
    """
    pass


class InvalidEnvironmentPathError(awsscripterException):
    """
    Error raised if the environment path string is invalid
    """
    pass


class ConfigItemNotFoundError(awsscripterException):
    """
    Error raised if a necessary config item has not been provided
    """
    pass


class UnsupportedTemplateFileTypeError(awsscripterException):
    """
    Error raised if an unsupported template file type is used.
    """
    pass


class TemplateawsscripterHandlerError(awsscripterException):
    """
    Error raised if awsscripter_handler() is not defined correctly in the template.
    """
    pass


class DependencyStackNotLaunchedError(awsscripterException):
    """
    Error raised when a dependency stack has not been launched
    """
    pass


class DependencyStackMissingOutputError(awsscripterException):
    """
    Error raised if a dependency stack does not have the correct outputs.
    """
    pass


class CircularDependenciesError(awsscripterException):
    """
    Error raised if there are circular dependencies
    """
    pass


class UnknownStackStatusError(awsscripterException):
    """
    Error raised if an unknown stack status is received.
    """
    pass


class UnknownAuditStatusError(awsscripterException):
    """
    Error raised if an unknown stack status is received.
    """
    pass

class RetryLimitExceededError(awsscripterException):
    """
    Error raised if the request limit is exceeded.
    """
    pass


class UnknownHookTypeError(awsscripterException):
    """
    Error raised if an unrecognised hook type is received.
    """


class VersionIncompatibleError(awsscripterException):
    """
    Error raised if configuration incompatible with running version.
    """
    pass


class ProtectedStackError(awsscripterException):
    """
    Error raised upon execution of an action under active protection
    """
    pass


class UnknownStackChangeSetStatusError(awsscripterException):
    """
    Error raised if an unknown stack change set status is received.
    """
    pass


class InvalidHookArgumentTypeError(awsscripterException):
    """
    Error raised if a hook's argument type is invalid.
    """
    pass


class InvalidHookArgumentSyntaxError(awsscripterException):
    """
    Error raised if a hook's argument syntax is invalid.
    """
    pass


class InvalidHookArgumentValueError(awsscripterException):
    """
    Error raised if a hook's argument value is invalid.
    """
    pass


class CannotUpdateFailedStackError(awsscripterException):
    """
    Error raised when a failed stack is updated.
    """
    pass


class StackDoesNotExistError(awsscripterException):
    """
    Error raised when a stack does not exist.
    """
    pass


class ConfigFileNotFoundError(awsscripterException):
    """
    Error raised when a config file does not exist.
    """
    pass


class EnvironmentNotFoundError(awsscripterException):
    """
    Error raised when a environment does not exist.
    """
    pass



import click

from awsscripter.stack.helpers import catch_exceptions, confirmation

@click.command(name="HIPPA")
@click.argument("level")
@click.pass_context
@catch_exceptions
def audit_hippa_command(ctx, level):
    """
    Depending on level , it will perform the CISP audit, by default the level is FULL
    """
    print("Auditing with level  " + level)
